-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

--External Pressure control for 4 Wheels vehicle, simultaneous inflate and deflate on both axles possible.
--Wheel groups must be named "front" and "rear1" in the jbeam.

--Modifications by Olivier ADLER - Olivier80


local M = {}
M.type = "auxiliary"

local max = math.max

local wheelGroups = {}
local activeGroups = {}
local inflateMaxFlow
local deflateMaxFlow
--local pEnv = 101325
--Mod to allow relative tire pressure < atmospheric pressure
local pEnv = 60000
local FrontGroupPressureChange = 0
local RearGroupPressureChange = 0


local function updateGFX(dt)
  
  --Mod to allow pressure less than atmospheric pressure
  --pEnv = obj:getEnvPressure()
  local activeGroupAvgPressure = 0
  local activeGroupCount = 0

  for groupName, wheelGroup in pairs(wheelGroups) do
    local groupPressure = 0
    for _, wheelData in pairs(wheelGroup.wheelData) do
      wheelData.isWheelBrokenOff = wheels.wheels[wheelData.wheelId].isBroken
      wheelData.currentPressure = obj:getGroupPressure(wheelData.pressureGroupId)
      groupPressure = groupPressure + (wheelData.isWheelBrokenOff and 0 or wheelData.currentPressure)
    end
    groupPressure = groupPressure / wheelGroup.wheelCount
    local preSyncGroupPressure = groupPressure --save the current group pressure for purposes of pressure equalization between wheels of a single axle
    groupPressure = 0 --reset the avg pressure so that we can re-calculate it with the changed pressures after equalization

    for _, wheelData in pairs(wheelGroup.wheelData) do
      local pressureDiff = preSyncGroupPressure - wheelData.currentPressure
      local pressureChange = clamp(pressureDiff * dt, -wheelGroup.groupEqualizeFlowLimit, wheelGroup.groupEqualizeFlowLimit)
      wheelData.currentPressure = wheelData.isWheelBrokenOff and wheelData.currentPressure or max(wheelData.currentPressure + pressureChange, pEnv)
	  --limit lowest pressure to environment pressure

      obj:setGroupPressure(wheelData.pressureGroupId, wheelData.currentPressure)
      groupPressure = groupPressure + wheelData.currentPressure
      --debug
    end
    groupPressure = groupPressure / wheelGroup.wheelCount --calculate adjusted avg again for equalization between active circuits

    if activeGroups[groupName] then
      activeGroupAvgPressure = activeGroupAvgPressure + groupPressure
      activeGroupCount = activeGroupCount + 1
    end
  end

  activeGroupAvgPressure = activeGroupCount > 0 and (activeGroupAvgPressure / activeGroupCount) or pEnv -- calculate avg between all active circuits

    local FrontpressureChangeInflateDeflate = FrontGroupPressureChange * dt
	local RearpressureChangeInflateDeflate = RearGroupPressureChange * dt
  for groupName, isActive in pairs(activeGroups) do
    local wheelGroup = wheelGroups[groupName]
    if isActive then
      for _, wheelData in pairs(wheelGroup.wheelData) do
        local pressureDiff = activeGroupAvgPressure - wheelData.currentPressure
        local pressureChangeEqualize = clamp(pressureDiff * dt, -wheelGroup.groupEqualizeFlowLimit * dt, wheelGroup.groupEqualizeFlowLimit * dt)

		local FrontlimitAdjustedPressureChangeInflateDeflate = FrontpressureChangeInflateDeflate
		local RearlimitAdjustedPressureChangeInflateDeflate = RearpressureChangeInflateDeflate
		
		if groupName == "front" then
			if (FrontpressureChangeInflateDeflate > 0 and wheelData.currentPressure >= wheelGroup.maxPressure) or (FrontpressureChangeInflateDeflate < 0 and wheelData.currentPressure <= wheelGroup.minPressure) then
			  FrontlimitAdjustedPressureChangeInflateDeflate = 0
			end
			wheelData.currentPressure = max(wheelData.currentPressure + pressureChangeEqualize + FrontlimitAdjustedPressureChangeInflateDeflate, pEnv) --limit lowest pressure to environment pressure
		end
		if groupName == "rear1" then
			if (RearpressureChangeInflateDeflate > 0 and wheelData.currentPressure >= wheelGroup.maxPressure) or (RearpressureChangeInflateDeflate < 0 and wheelData.currentPressure <= wheelGroup.minPressure) then
				RearlimitAdjustedPressureChangeInflateDeflate = 0
			end		
		wheelData.currentPressure = max(wheelData.currentPressure + pressureChangeEqualize + RearlimitAdjustedPressureChangeInflateDeflate, pEnv) --limit lowest pressure to environment pressure
		end
        obj:setGroupPressure(wheelData.pressureGroupId, wheelData.currentPressure)
      end
    end
  end

  --controller buttons input
	if electrics.values.F_deflate == 1 and electrics.values.F_inflate == 0 then
		gui.message("Deflate Front", 2, "vehicle.tire.pressureF")
		activeGroups.front = true
		FrontGroupPressureChange = -deflateMaxFlow
	end
	if electrics.values.F_inflate == 1 and electrics.values.F_deflate == 0 then
		gui.message("Inflate Front", 2, "vehicle.tire.pressureF")
		activeGroups.front = true
		FrontGroupPressureChange = inflateMaxFlow
	end
	if electrics.values.R_deflate == 1 and electrics.values.R_inflate == 0 then
		gui.message("Deflate Rear", 2, "vehicle.tire.pressureR")
		activeGroups.rear1 = true
		RearGroupPressureChange = -deflateMaxFlow
	end
	if electrics.values.R_inflate == 1 and electrics.values.R_deflate == 0 then
		gui.message("Inflate Rear", 2, "vehicle.tire.pressureR")
		activeGroups.rear1 = true
		RearGroupPressureChange = inflateMaxFlow		
	end
	if (electrics.values.F_inflate == 0 and electrics.values.F_deflate == 0) or (electrics.values.F_inflate == 1 and electrics.values.F_deflate == 1) then
		activeGroups.front = false
		FrontGroupPressureChange = 0
	end
	if (electrics.values.R_inflate == 0 and electrics.values.R_deflate == 0) or (electrics.values.R_inflate == 1 and electrics.values.R_deflate == 1) then
		activeGroups.rear1 = false
		RearGroupPressureChange = 0
	end

end

local function reset(jbeamData)
  for groupName, wheelGroup in pairs(wheelGroups) do
    activeGroups[groupName] = false
    for _, wheelData in pairs(wheelGroup.wheelData) do
      wheelData.currentPressure = 0
      wheelData.isWheelBrokenOff = false
    end
  end

end

-- local function initSounds(jbeamData)
-- end

local function init(jbeamData)
end

local function initSecondStage(jbeamData)
  local mode = jbeamData.mode or "manualControl"
  if mode == "manualControl" then
    wheelGroups = {}
    activeGroups = {}
    inflateMaxFlow = jbeamData.inflateMaxFlow or 10
    deflateMaxFlow = jbeamData.deflateMaxFlow or 10
    local wheelGroupData = tableFromHeaderTable(jbeamData.wheelGroups or {})
    for _, wheelGroup in pairs(wheelGroupData) do
      local groupName = wheelGroup.groupName
      local uiName = wheelGroup.uiName
      local uiIcon = wheelGroup.uiIcon
      local wheelNames = wheelGroup.wheelNames
      local maxPressure = wheelGroup.maxPressure
      local minPressure = wheelGroup.minPressure
      local groupEqualizeFlowLimit = wheelGroup.groupEqualizeFlowLimit
      if groupName and uiName and wheelNames and #wheelNames > 0 and maxPressure and minPressure and groupEqualizeFlowLimit then
        local wheelData = {}
        for _, wheelName in pairs(wheelNames) do
          local wheelId = wheels.wheelIDs[wheelName]
          if wheelId then
            local pressureGroupId = wheels.wheels[wheelId].pressureGroupId
            if pressureGroupId then
              wheelData[wheelName] = {
                pressureGroupId = pressureGroupId,
                wheelId = wheelId,
                isWheelBrokenOff = false,
                currentPressure = obj:getGroupPressure(pressureGroupId)
              }
            else
              log("E", "tirePressureControl.initSecondStage", string.format("Can't find pressure group id for wheel %q", wheelName))
            end
          else
            log("E", "tirePressureControl.initSecondStage", string.format("Can't find wheel id for wheel %q", wheelName))
          end
        end

        if not tableIsEmpty(wheelData) then
          wheelGroups[groupName] = {
            uiName = uiName,
            uiIcon = uiIcon,
            wheelNames = wheelNames,
            wheelData = wheelData,
            wheelCount = tableSize(wheelData),
            maxPressure = maxPressure,
            minPressure = minPressure,
            groupEqualizeFlowLimit = groupEqualizeFlowLimit,
            isActiveElectricsName = M.name .. "_" .. groupName .. "_isActive"
          }
          activeGroups[groupName] = false
        else
          log("D", "tirePressureControl.initSecondStage", "No valid wheel data provided for group: " .. (groupName or "nil"))
        end
      else
        log("E", "tirePressureControl.initSecondStage", "Invalid jbeam data provided")
        log("E", "tirePressureControl.initSecondStage", dumps(jbeamData))
      end
    end
  end

  electrics.values.F_inflate = 0
  electrics.values.F_deflate = 0
  electrics.values.R_inflate = 0
  electrics.values.R_deflate = 0
end

M.init = init
M.initSecondStage = initSecondStage
--M.initSounds = initSounds
M.reset = reset
--M.resetSounds = resetSounds

M.updateGFX = updateGFX

return M
